@@ -9,19 +9,20 @@ use wgpu::{
 #[allow(unused_imports)]
 use crate::format_sso;
 use crate::{
-    resources::STARTING_2D_TEXTURES, util::typedefs::FastHashMap, LimitType, RendererInitializationError, RendererMode,
+    managers::STARTING_2D_TEXTURES, util::typedefs::FastHashMap, LimitType, RendererInitializationError,
+    RendererProfile,
 };
 
 /// Largest uniform buffer binding needed to run rend3.
 pub const MAX_UNIFORM_BUFFER_BINDING_SIZE: BufferAddress = 1024;
 
-/// Features required to run in gpu-mode.
-pub const GPU_REQUIRED_FEATURES: Features = {
+/// Features required to run in the GpuDriven profile.
+pub const GPU_DRIVEN_REQUIRED_FEATURES: Features = {
     // We need to do this whole bits thing to make this const as OpOr isn't const
     Features::from_bits_truncate(
         Features::PUSH_CONSTANTS.bits()
             | Features::TEXTURE_COMPRESSION_BC.bits()
-            | Features::DEPTH_CLAMPING.bits()
+            | Features::DEPTH_CLIP_CONTROL.bits()
             | Features::TEXTURE_BINDING_ARRAY.bits()
             | Features::SAMPLED_TEXTURE_AND_STORAGE_BUFFER_ARRAY_NON_UNIFORM_INDEXING.bits()
             | Features::PARTIALLY_BOUND_BINDING_ARRAY.bits()
@@ -32,22 +33,24 @@ pub const GPU_REQUIRED_FEATURES: Features = {
     )
 };
 
-/// Features required to run in cpu-mode.
-pub const CPU_REQUIRED_FEATURES: Features = Features::from_bits_truncate(Features::DEPTH_CLAMPING.bits());
+/// Features required to run in the GpuDriven profile.
+pub const CPU_DRIVEN_REQUIRED_FEATURES: Features = Features::from_bits_truncate(0);
 
 /// Features that rend3 can use if it they are available, but we don't require.
 pub const OPTIONAL_FEATURES: Features = Features::from_bits_truncate(
-    Features::TEXTURE_COMPRESSION_BC.bits()
+    Features::DEPTH_CLIP_CONTROL.bits()
+        | Features::TEXTURE_COMPRESSION_BC.bits()
         | Features::TEXTURE_COMPRESSION_ETC2.bits()
         | Features::TEXTURE_COMPRESSION_ASTC_LDR.bits()
         | Features::TIMESTAMP_QUERY.bits(),
 );
 
-/// Check that all required features for a given mode are present in the feature set given.
-pub fn check_features(mode: RendererMode, device: Features) -> Result<Features, RendererInitializationError> {
-    let required = match mode {
-        RendererMode::GPUPowered => GPU_REQUIRED_FEATURES,
-        RendererMode::CPUPowered => CPU_REQUIRED_FEATURES,
+/// Check that all required features for a given profile are present in the feature
+/// set given.
+pub fn check_features(profile: RendererProfile, device: Features) -> Result<Features, RendererInitializationError> {
+    let required = match profile {
+        RendererProfile::GpuDriven => GPU_DRIVEN_REQUIRED_FEATURES,
+        RendererProfile::CpuDriven => CPU_DRIVEN_REQUIRED_FEATURES,
     };
     let optional = OPTIONAL_FEATURES & device;
     let missing = required - device;
@@ -58,13 +61,13 @@ pub fn check_features(mode: RendererMode, device: Features) -> Result<Features,
     }
 }
 
-/// Limits required to run in gpu-mode.
+/// Limits required to run in the GpuDriven profile.
 pub const GPU_REQUIRED_LIMITS: Limits = Limits {
     max_texture_dimension_1d: 2048,
     max_texture_dimension_2d: 2048,
     max_texture_dimension_3d: 512,
-    max_texture_array_layers: 512,
-    max_bind_groups: 8,
+    max_texture_array_layers: 256,
+    max_bind_groups: 6,
     max_dynamic_uniform_buffers_per_pipeline_layout: 0,
     max_dynamic_storage_buffers_per_pipeline_layout: 0,
     max_sampled_textures_per_shader_stage: STARTING_2D_TEXTURES as _,
@@ -80,15 +83,22 @@ pub const GPU_REQUIRED_LIMITS: Limits = Limits {
     max_push_constant_size: 128,
     min_uniform_buffer_offset_alignment: 256,
     min_storage_buffer_offset_alignment: 256,
+    max_inter_stage_shader_components: 60,
+    max_compute_workgroup_storage_size: 16352,
+    max_compute_invocations_per_workgroup: 256,
+    max_compute_workgroup_size_x: 256,
+    max_compute_workgroup_size_y: 256,
+    max_compute_workgroup_size_z: 64,
+    max_compute_workgroups_per_dimension: 65535,
 };
 
-/// Limits required to run in cpu-mode.
+/// Limits required to run in the CpuDriven profile.
 pub const CPU_REQUIRED_LIMITS: Limits = Limits {
     max_texture_dimension_1d: 2048,
     max_texture_dimension_2d: 2048,
     max_texture_dimension_3d: 512,
-    max_texture_array_layers: 512,
-    max_bind_groups: 8,
+    max_texture_array_layers: 256,
+    max_bind_groups: 4,
     max_dynamic_uniform_buffers_per_pipeline_layout: 0,
     max_dynamic_storage_buffers_per_pipeline_layout: 0,
     max_sampled_textures_per_shader_stage: 10,
@@ -104,6 +114,13 @@ pub const CPU_REQUIRED_LIMITS: Limits = Limits {
     max_push_constant_size: 0,
     min_uniform_buffer_offset_alignment: 256,
     min_storage_buffer_offset_alignment: 256,
+    max_inter_stage_shader_components: 60,
+    max_compute_workgroup_storage_size: 16352,
+    max_compute_invocations_per_workgroup: 256,
+    max_compute_workgroup_size_x: 256,
+    max_compute_workgroup_size_y: 256,
+    max_compute_workgroup_size_z: 64,
+    max_compute_workgroups_per_dimension: 65535,
 };
 
 fn check_limit_unlimited(d: u32, r: u32, ty: LimitType) -> Result<u32, RendererInitializationError> {
@@ -130,11 +147,12 @@ fn check_limit_low(d: u32, r: u32, ty: LimitType) -> Result<u32, RendererInitial
     }
 }
 
-/// Check that all required limits for a given mode are present in the given limit set.
-pub fn check_limits(mode: RendererMode, device_limits: &Limits) -> Result<Limits, RendererInitializationError> {
-    let required_limits = match mode {
-        RendererMode::GPUPowered => GPU_REQUIRED_LIMITS,
-        RendererMode::CPUPowered => CPU_REQUIRED_LIMITS,
+/// Check that all required limits for a given profile are present in the given
+/// limit set.
+pub fn check_limits(profile: RendererProfile, device_limits: &Limits) -> Result<Limits, RendererInitializationError> {
+    let required_limits = match profile {
+        RendererProfile::GpuDriven => GPU_REQUIRED_LIMITS,
+        RendererProfile::CpuDriven => CPU_REQUIRED_LIMITS,
     };
 
     Ok(Limits {
@@ -238,6 +256,41 @@ pub fn check_limits(mode: RendererMode, device_limits: &Limits) -> Result<Limits
             required_limits.min_uniform_buffer_offset_alignment,
             LimitType::UniformBufferBindingAlignment,
         )?,
+        max_inter_stage_shader_components: check_limit_unlimited(
+            device_limits.max_inter_stage_shader_components,
+            required_limits.max_inter_stage_shader_components,
+            LimitType::MaxInterStageShaderComponents,
+        )?,
+        max_compute_workgroup_storage_size: check_limit_unlimited(
+            device_limits.max_compute_workgroup_storage_size,
+            required_limits.max_compute_workgroup_storage_size,
+            LimitType::MaxComputeWorkgroupStorageSize,
+        )?,
+        max_compute_invocations_per_workgroup: check_limit_unlimited(
+            device_limits.max_compute_invocations_per_workgroup,
+            required_limits.max_compute_invocations_per_workgroup,
+            LimitType::MaxComputeInvocationsPerWorkgroup,
+        )?,
+        max_compute_workgroup_size_x: check_limit_unlimited(
+            device_limits.max_compute_workgroup_size_x,
+            required_limits.max_compute_workgroup_size_x,
+            LimitType::MaxComputeWorkgroupSizeX,
+        )?,
+        max_compute_workgroup_size_y: check_limit_unlimited(
+            device_limits.max_compute_workgroup_size_y,
+            required_limits.max_compute_workgroup_size_y,
+            LimitType::MaxComputeWorkgroupSizeY,
+        )?,
+        max_compute_workgroup_size_z: check_limit_unlimited(
+            device_limits.max_compute_workgroup_size_z,
+            required_limits.max_compute_workgroup_size_z,
+            LimitType::MaxComputeWorkgroupSizeZ,
+        )?,
+        max_compute_workgroups_per_dimension: check_limit_unlimited(
+            device_limits.max_compute_workgroups_per_dimension,
+            required_limits.max_compute_workgroups_per_dimension,
+            LimitType::MaxComputeWorkgroupsPerDimension,
+        )?,
     })
 }
 
@@ -247,7 +300,7 @@ pub struct PotentialAdapter<T> {
     pub info: ExtendedAdapterInfo,
     pub features: Features,
     pub limits: Limits,
-    pub mode: RendererMode,
+    pub profile: RendererProfile,
 }
 impl<T> PotentialAdapter<T> {
     pub fn new(
@@ -255,20 +308,20 @@ impl<T> PotentialAdapter<T> {
         inner_info: AdapterInfo,
         inner_limits: Limits,
         inner_features: Features,
-        desired_mode: Option<RendererMode>,
+        desired_profile: Option<RendererProfile>,
     ) -> Result<Self, RendererInitializationError> {
         let info = ExtendedAdapterInfo::from(inner_info);
 
-        let mut features = check_features(RendererMode::GPUPowered, inner_features);
-        let mut limits = check_limits(RendererMode::GPUPowered, &inner_limits);
-        let mut mode = RendererMode::GPUPowered;
+        let mut features = check_features(RendererProfile::GpuDriven, inner_features);
+        let mut limits = check_limits(RendererProfile::GpuDriven, &inner_limits);
+        let mut profile = RendererProfile::GpuDriven;
 
-        if (features.is_err() || limits.is_err() || desired_mode == Some(RendererMode::CPUPowered))
-            && desired_mode != Some(RendererMode::GPUPowered)
+        if (features.is_err() || limits.is_err() || desired_profile == Some(RendererProfile::CpuDriven))
+            && desired_profile != Some(RendererProfile::GpuDriven)
         {
-            features = check_features(RendererMode::CPUPowered, inner_features);
-            limits = check_limits(RendererMode::CPUPowered, &inner_limits);
-            mode = RendererMode::CPUPowered;
+            features = check_features(RendererProfile::CpuDriven, inner_features);
+            limits = check_limits(RendererProfile::CpuDriven, &inner_limits);
+            profile = RendererProfile::CpuDriven;
         }
 
         Ok(PotentialAdapter {
@@ -276,7 +329,7 @@ impl<T> PotentialAdapter<T> {
             info,
             features: features?,
             limits: limits?,
-            mode,
+            profile,
         })
     }
 }
@@ -334,23 +387,34 @@ impl From<AdapterInfo> for ExtendedAdapterInfo {
 /// Container for Instance/Adapter/Device/Queue etc.
 ///
 /// Create these yourself, or call [`create_iad`].
+#[derive(Clone)]
 pub struct InstanceAdapterDevice {
     pub instance: Arc<Instance>,
     pub adapter: Arc<Adapter>,
     pub device: Arc<Device>,
     pub queue: Arc<Queue>,
-    pub mode: RendererMode,
+    pub profile: RendererProfile,
     pub info: ExtendedAdapterInfo,
 }
 
-/// Creates an Instance/Adapter/Device/Queue using the given choices. Tries to get the best combination.
+/// Creates an Instance/Adapter/Device/Queue using the given choices. Tries to
+/// get the best combination.
+///
+/// **NOTE:** Some adapters will not advertise all of its supported features.
+/// The `additional_features` parameter can be used to explicitly request
+/// additional features during device creation.
 pub async fn create_iad(
     desired_backend: Option<Backend>,
     desired_device: Option<String>,
-    desired_mode: Option<RendererMode>,
+    desired_profile: Option<RendererProfile>,
+    additional_features: Option<Features>,
 ) -> Result<InstanceAdapterDevice, RendererInitializationError> {
     profiling::scope!("create_iad");
+    #[cfg(not(target_arch = "wasm32"))]
     let backend_bits = Backends::VULKAN | Backends::DX12 | Backends::DX11 | Backends::METAL | Backends::GL;
+    #[cfg(target_arch = "wasm32")]
+    let backend_bits = Backends::BROWSER_WEBGPU;
+    #[cfg(not(target_arch = "wasm32"))]
     let default_backend_order = [
         Backend::Vulkan,
         Backend::Metal,
@@ -358,23 +422,35 @@ pub async fn create_iad(
         Backend::Dx11,
         Backend::Gl,
     ];
+    #[cfg(target_arch = "wasm32")]
+    let default_backend_order = [Backend::BrowserWebGpu];
 
     let instance = Instance::new(backend_bits);
 
-    let mut valid_adapters = FastHashMap::default();
+    let mut valid_adapters = FastHashMap::<Backend, ArrayVec<PotentialAdapter<Adapter>, 4>>::default();
 
     for backend in &default_backend_order {
         profiling::scope!("enumerating backend", &format_sso!("{:?}", backend));
+        #[cfg(not(target_arch = "wasm32"))]
         let adapters = instance.enumerate_adapters(Backends::from(*backend));
+        #[cfg(target_arch = "wasm32")]
+        let adapters = instance
+            .request_adapter(&wgpu::RequestAdapterOptions {
+                power_preference: wgpu::PowerPreference::HighPerformance,
+                force_fallback_adapter: false,
+                compatible_surface: None,
+            })
+            .await
+            .into_iter();
 
         let mut potential_adapters = ArrayVec::<PotentialAdapter<Adapter>, 4>::new();
         for (idx, adapter) in adapters.enumerate() {
             let info = adapter.get_info();
             let limits = adapter.limits();
             let features = adapter.features();
-            let potential = PotentialAdapter::new(adapter, info, limits, features, desired_mode);
+            let potential = PotentialAdapter::new(adapter, info, limits, features, desired_profile);
 
-            log::debug!(
+            log::info!(
                 "{:?} Adapter {}: {:#?}",
                 backend,
                 idx,
@@ -391,7 +467,7 @@ pub async fn create_iad(
             };
 
             if let (Ok(potential), true) = (potential, desired) {
-                log::debug!("Adapter usable in {:?} mode", potential.mode);
+                log::debug!("Adapter usable in the {:?} profile", potential.profile);
                 potential_adapters.push(potential)
             } else {
                 log::debug!("Adapter not usable");
@@ -426,14 +502,16 @@ pub async fn create_iad(
             log::debug!("Chosen backend: {:?}", backend);
             log::debug!("Chosen features: {:#?}", adapter.features);
             log::debug!("Chosen limits: {:#?}", adapter.limits);
-            log::debug!("Chosen mode: {:#?}", adapter.mode);
+            log::debug!("Chosen profile: {:#?}", adapter.profile);
 
             let (device, queue) = adapter
                 .inner
                 .request_device(
                     &DeviceDescriptor {
                         label: None,
-                        features: adapter.features,
+                        features: adapter
+                            .features
+                            .union(additional_features.unwrap_or_else(Features::empty)),
                         limits: adapter.limits,
                     },
                     None,
@@ -446,7 +524,7 @@ pub async fn create_iad(
                 adapter: Arc::new(adapter.inner),
                 device: Arc::new(device),
                 queue: Arc::new(queue),
-                mode: adapter.mode,
+                profile: adapter.profile,
                 info: adapter.info,
             });
         }
