@@ -1,29 +1,31 @@
+//! Mipmap generation tools.
+
 use std::num::NonZeroU32;
 
 use arrayvec::ArrayVec;
-use parking_lot::{Mutex, RwLock};
+use parking_lot::RwLock;
 use rend3_types::TextureFormat;
 use wgpu::{
     AddressMode, BindGroup, BindGroupLayout, BindGroupLayoutDescriptor, BindGroupLayoutEntry, BindingType, Color,
     ColorTargetState, ColorWrites, CommandEncoder, Device, FilterMode, FragmentState, FrontFace, LoadOp,
     MultisampleState, Operations, PipelineLayout, PipelineLayoutDescriptor, PolygonMode, PrimitiveState,
     PrimitiveTopology, RenderPassColorAttachment, RenderPassDescriptor, RenderPipeline, RenderPipelineDescriptor,
-    SamplerDescriptor, ShaderModule, ShaderStages, Texture, TextureDescriptor, TextureSampleType,
+    SamplerBindingType, SamplerDescriptor, ShaderModule, ShaderStages, Texture, TextureDescriptor, TextureSampleType,
     TextureViewDescriptor, TextureViewDimension, VertexState,
 };
-use wgpu_profiler::GpuProfiler;
 
 use crate::{
     format_sso,
     util::{bind_merge::BindGroupBuilder, typedefs::FastHashMap},
 };
 
+/// Generator for mipmaps.
 pub struct MipmapGenerator {
-    pub texture_bgl: BindGroupLayout,
-    pub sampler_bg: BindGroup,
-    pub sm: ShaderModule,
-    pub pll: PipelineLayout,
-    pub pipelines: RwLock<FastHashMap<TextureFormat, RenderPipeline>>,
+    texture_bgl: BindGroupLayout,
+    sampler_bg: BindGroup,
+    sm: ShaderModule,
+    pll: PipelineLayout,
+    pipelines: RwLock<FastHashMap<TextureFormat, RenderPipeline>>,
 }
 
 impl MipmapGenerator {
@@ -49,10 +51,7 @@ impl MipmapGenerator {
             entries: &[BindGroupLayoutEntry {
                 binding: 0,
                 visibility: ShaderStages::FRAGMENT,
-                ty: BindingType::Sampler {
-                    filtering: true,
-                    comparison: false,
-                },
+                ty: BindingType::Sampler(SamplerBindingType::Filtering),
                 count: None,
             }],
         });
@@ -65,18 +64,20 @@ impl MipmapGenerator {
             mag_filter: FilterMode::Linear,
             min_filter: FilterMode::Linear,
             mipmap_filter: FilterMode::Linear,
-            lod_min_clamp: -100.0,
+            lod_min_clamp: 0.0,
             lod_max_clamp: 100.0,
             compare: None,
             anisotropy_clamp: None,
             border_color: None,
         });
 
-        let sampler_bg = BindGroupBuilder::new(Some("mipmap generator sampler bg"))
-            .with_sampler(&sampler)
-            .build(device, &sampler_bgl);
+        let sampler_bg = BindGroupBuilder::new().append_sampler(&sampler).build(
+            device,
+            Some("mipmap generator sampler bg"),
+            &sampler_bgl,
+        );
 
-        let sm = device.create_shader_module(&wgpu::include_wgsl!("mipmap.wgsl"));
+        let sm = device.create_shader_module(&wgpu::include_wgsl!("../../shaders/mipmap.wgsl"));
 
         let pll = device.create_pipeline_layout(&PipelineLayoutDescriptor {
             label: Some("mipmap generator pipeline layout"),
@@ -119,7 +120,7 @@ impl MipmapGenerator {
                 strip_index_format: None,
                 front_face: FrontFace::Cw,
                 cull_mode: None,
-                clamp_depth: false,
+                unclipped_depth: false,
                 polygon_mode: PolygonMode::Fill,
                 conservative: false,
             },
@@ -134,14 +135,13 @@ impl MipmapGenerator {
                     write_mask: ColorWrites::all(),
                 }],
             }),
+            multiview: None,
         })
     }
 
     pub fn generate_mipmaps(
         &self,
         device: &Device,
-        // TODO: allow profiler to run asynchronously
-        _profiler: &Mutex<GpuProfiler>,
         encoder: &mut CommandEncoder,
         texture: &Texture,
         desc: &TextureDescriptor,
@@ -185,9 +185,11 @@ impl MipmapGenerator {
             profiling::scope!(&_dst_label);
             // profiler.lock().begin_scope(&dst_label, encoder, device);
 
-            let bg = BindGroupBuilder::new(Some(&src_label))
-                .with_texture_view(src_view)
-                .build(device, &self.texture_bgl);
+            let bg = BindGroupBuilder::new().append_texture_view(src_view).build(
+                device,
+                Some(&src_label),
+                &self.texture_bgl,
+            );
 
             let mut rpass = encoder.begin_render_pass(&RenderPassDescriptor {
                 label: None,
